--[[
    Aaron MacDonald
    August 28, 2010
]]

require('graphics')
require('animation')
require('constants')
require('utility')
require('levels')

local WALK_SPEED = 5

actor = {}

function actor.diagonalSpeed(speed)
    return (speed * speed) / math.sqrt((speed * speed) * 2)
end

----------------------------------------------------------------------

function walkQuads(actWidth, actHeight, sheetWidth, sheetHeight, y)
    local quads = {
      love.graphics.newQuad(actWidth, y, actWidth, actHeight,
        sheetWidth, sheetHeight),
      love.graphics.newQuad(0, y, actWidth, actHeight,
        sheetWidth, sheetHeight),
      love.graphics.newQuad(actWidth * 2, y, actWidth, actHeight,
        sheetWidth, sheetHeight)
    }

    return quads
end

function actor.buildType(baseType)
    local spriteSheet = graphics.loadImage(baseType.spriteSheet)

    local actWidth = baseType.spriteWidth
    local actHeight = baseType.spriteHeight
    local sheetWidth = spriteSheet:getWidth()
    local sheetHeight = spriteSheet:getHeight()

    local idleQuads = {
      [constants.SOUTH] = love.graphics.newQuad(0, 0,
        actWidth, actHeight, sheetWidth, sheetHeight),
      [constants.NORTH] = love.graphics.newQuad( 0, actHeight,
        actWidth, actHeight, sheetWidth, sheetHeight),
      [constants.WEST] = love.graphics.newQuad(0, actHeight * 2,
        actWidth, actHeight, sheetWidth, sheetHeight),
      [constants.EAST] = love.graphics.newQuad(0, actHeight * 3,
        actWidth, actHeight, sheetWidth, sheetHeight)
    }

    local walkQuads = {
      [constants.SOUTH] = walkQuads(actWidth, actHeight,
        sheetWidth, sheetHeight, 0),
      [constants.NORTH] = walkQuads(actWidth, actHeight,
        sheetWidth, sheetHeight, actHeight),
      [constants.WEST] = walkQuads(actWidth, actHeight,
        sheetWidth, sheetHeight, actHeight * 2),
      [constants.EAST] = walkQuads(actWidth, actHeight,
        sheetWidth, sheetHeight, actHeight * 3)
    }

    local diagonalSpeed = actor.diagonalSpeed(baseType.speed)

    local bodyOffsetY = baseType.bodyHeight - actHeight

    return { spriteSheet = spriteSheet,

             idleQuads = idleQuads,
             walkQuads = walkQuads,

             spriteOffsetX = -actWidth / 2,
             spriteOffsetY = -baseType.offsetY,

             spriteHeight = actHeight, -- For z-ordering

             groundRadius = baseType.groundRadius,

             bodyWidth = baseType.bodyWidth,
             bodyHeight = baseType.bodyHeight,
             bodyOffsetY = bodyOffsetY,

             speed = baseType.speed,
             diagonalSpeed = diagonalSpeed }
end

----------------------------------------------------------------------

local function buildPhysics(act, level, x, y, isPlayer)
    act.body = love.physics.newBody(level.physicsWorld, x, y, 1, 0)

    local groundCircle = love.physics.newCircleShape(act.body,
      0, 0, act.type.groundRadius)
    groundCircle:setRestitution(0)
    groundCircle:setCategory(constants.CAT_ACTOR_GROUND)
    groundCircle:setMask(constants.CAT_ACTOR_GROUND,
      constants.CAT_ACTOR_BODY) -- Only collide with terrain

    local bodyRect = love.physics.newRectangleShape(act.body,
      0, act.type.bodyOffsetY,
      act.type.bodyWidth, act.type.bodyHeight)
    bodyRect:setRestitution(0)
    bodyRect:setCategory(constants.CAT_ACTOR_BODY)
    -- Only collide with actor bodies
    bodyRect:setMask(constants.CAT_WALL, constants.CAT_ACTOR_GROUND)

    if isPlayer then
        utility.setShapeData(groundCircle, constants.PLAYER, act)
        utility.setShapeData(bodyRect, constants.PLAYER, act)
    end

    act.shapes = {groundCircle, bodyRect}

    act.body:setBullet(true) -- I'm a sucker for
                             -- continuous collision detection
end

function actor.newActor(level, actorType, x, y, isPlayer, facing)
    local act = {
        type = actorType,
        facing = facing or constants.SOUTH,
        moving = false,

        walkAnims = {
          [constants.SOUTH] = animation.newAnim(actorType.spriteSheet,
            actorType.walkQuads[constants.SOUTH], WALK_SPEED),
          [constants.NORTH] = animation.newAnim(actorType.spriteSheet,
            actorType.walkQuads[constants.NORTH], WALK_SPEED),
          [constants.WEST] = animation.newAnim(actorType.spriteSheet,
            actorType.walkQuads[constants.WEST], WALK_SPEED),
          [constants.EAST] = animation.newAnim(actorType.spriteSheet,
            actorType.walkQuads[constants.EAST], WALK_SPEED)
        }
    }

    buildPhysics(act, level, x, y, isPlayer)

    return act
end

function actor.changeLevel(act, level, x, y, isPlayer)
    buildPhysics(act, level, x, y, isPlayer)
end

-- function actor.disableCollisions(act, category)
--     for _, s in ipairs(act.shapes)
--     do
--         s:setMask(category)
--     end
-- end
-- 
-- function actor.enableAllCollisions(act)
--     for _, s in ipairs(act.shapes)
--     do
--         s:setMask()
--     end
-- end

function actor.getX(act)
    return act.body:getX()
end

function actor.getY(act)
    return act.body:getY()
end

function actor.getRoundedX(act)
    return utility.round(actor.getX(act))
end

function actor.getRoundedY(act)
    return utility.round(actor.getY(act))
end

function actor.updateMotion(act, movementControls)
    local dx = 0
    local dy = 0

    if movementControls then
        local speed = act.type.speed
        local diagonalSpeed = act.type.diagonalSpeed

        -- Velocity

        if movementControls[constants.RIGHT] then
            dx = speed
        end
        if movementControls[constants.LEFT] then
            dx = dx - speed
        end

        if movementControls[constants.UP] then
            dy = -speed
        end
        if movementControls[constants.DOWN] then
            dy = dy + speed
        end

        -- Diagonal velocity

        if (dx ~= 0) and (dy ~= 0) then
            dx = diagonalSpeed * utility.sign(dx)
            dy = diagonalSpeed * utility.sign(dy)
        end

        -- Facing

        local newFacing

        if (dx > 0)
          and ((dy == 0) or (act.facing == constants.WEST)) then
            newFacing = constants.EAST
        end

        if (dx < 0)
          and ((dy == 0) or (act.facing == constants.EAST)) then
            newFacing = constants.WEST
        end

        if (dy > 0)
          and ((dx == 0) or (act.facing == constants.NORTH)) then
            newFacing = constants.SOUTH
        end

        if (dy < 0)
          and ((dx == 0) or (act.facing == constants.SOUTH)) then
            newFacing = constants.NORTH
        end

        if newFacing and (act.facing ~= newFacing) then
            animation.setFrame(act.walkAnims[act.facing])
            act.facing = newFacing
        end
    end

    act.moving = (dx ~= 0) or (dy ~= 0)

    act.body:wakeUp()
    act.body:setLinearVelocity(dx, dy)
end

function actor.draw(act)
    local x = actor.getRoundedX(act)
    local y = actor.getRoundedY(act)

    x = x + act.type.spriteOffsetX
    y = y + act.type.spriteOffsetY

    if act.moving then
        animation.draw(act.walkAnims[act.facing], x, y)
    else
        love.graphics.drawq(act.type.spriteSheet,
          act.type.idleQuads[act.facing], x, y)
    end
end