--[[
    Aaron MacDonald
    August 28, 2010
]]

require('constants')
require('graphics')
require('utility')

local LEVEL_DIR = 'levels/'
local LEVEL_EXT = '.lua'

local LAYER_GROUND = 'Ground'
local LAYER_OVER = 'Overhang'
local LAYER_ACTORS = 'Actors'
local LAYER_PORTALS = 'Portals'

local ACT_PLAYER = 'Player'

local TRANS_TILE_LABEL = 'tile'
local TRANS_LEVEL_LABEL = 'map'

levels = {}
levels.TILE_SIZE = 64
levels.TILE_HALF_SIZE = math.floor(levels.TILE_SIZE / 2)

levels.TILESETS = {}

----------------------------------------------------------------------

function levels.indexToCoords(index, width)
    -- Index starts at 1, because Lua's array indexing is bull#$*%.
    local realI = index - 1
    local y = math.floor(realI / width)
    local x = realI - (y * width)

    return x, y
end

--[[
function levels.coordsToIndex(x, y, width)
    -- Index starts at one
    return (y * width) + x + 1
end

function levels.maxIndex(level)
    return level.width * level.height
end
]]

function levels.inBounds(level, x, y)
    return (x >= 0) and (y >= 0)
      and (x < level.width) and (y < level.height)
end

function levels.pixel(coord)
    return coord * levels.TILE_SIZE
end

function levels.tile(pixel)
    return math.floor(pixel / levels.TILE_SIZE)
end

function levels.pixelWidth(level)
    return level.width * levels.TILE_SIZE
end

function levels.pixelHeight(level)
    return level.height * levels.TILE_SIZE
end

function levels.truePixelHeight(level)
    return levels.pixelHeight(level) - level.minY
end

function levels.offsetPixel(pixel)
    return pixel - (levels.tile(pixel) * levels.TILE_SIZE)
end

--[[
function levels.tilesInRect(level, rect, filter)
    local minX = levels.tile(rect.x)
    local minY = levels.tile(rect.y)
    local maxX = levels.tile(rect.x + rect.width - 1)
    local maxY = levels.tile(rect.y + rect.height - 1)

    if minX < 0 then
        minX = 0
    end

    if minY < 0 then
        minY = 0
    end

    if maxX >= level.width then
        maxX = level.width - 1
    end

    if maxY >= level.height then
        maxY = level.height - 1
    end

    local result = {}

    for x = minX, maxX do
        for y = minY, maxY do
            if (not filter) or filter(x, y) then
                table.insert(result, {x = x, y = y})
            end
        end
    end

    return result
end
]]

function levels.isWall(level, x, y)
    return level.tileset.walls[level.groundLayer[x][y]]
end

function levels.entranceCoords(level, portalID)
    if level.inPortals[portalID] then
        return level.inPortals[portalID].x,
          level.inPortals[portalID].y
    end
end

function levels.exitData(level, pixelX, pixelY)
    local result

    local x = math.floor(pixelX / levels.TILE_SIZE)
    local y = math.floor(pixelY / levels.TILE_SIZE)

    if level.outPortals[x] then
        result = level.outPortals[x][y]
    end

    return result
end

----------------------------------------------------------------------

local function makeWall(level, x, y, pixelCoords)
    local tx = x
    local ty = y

    if not pixelCoords then
        tx = (tx * levels.TILE_SIZE) + levels.TILE_HALF_SIZE
        ty = (ty * levels.TILE_SIZE) + levels.TILE_HALF_SIZE
    end

    local body = love.physics.newBody(level.physicsWorld,
      tx, ty, 0, 0)
    local shape = love.physics.newRectangleShape(
      body, 0, 0, levels.TILE_SIZE, levels.TILE_SIZE, 0 )
    shape:setRestitution(0)

    shape:setCategory(constants.CAT_WALL)

    return body, shape
end

local function makeBound(level, dir)
    -- A two-tile-thick wall along an edge

    assert( ((dir.x == 0) or (dir.y == 0)) and (dir.x ~= dir.y))

    local x, y, bWidth, bHeight

    if dir.x ~= 0 then
        if dir.x > 0 then
            x = levels.pixelWidth(level) + levels.TILE_SIZE
        else
            -- dir.x < 0
            x = -levels.TILE_SIZE
        end

        bHeight = levels.pixelWidth(level)
    else
        x = levels.pixelWidth(level) / 2
        bHeight = levels.TILE_SIZE * 2
    end

    if dir.y ~= 0 then
        if dir.y > 0 then
            y = level.minY - levels.TILE_SIZE
        else
            -- dir.y < 0
            y = levels.pixelHeight(level) + levels.TILE_SIZE
        end

        bWidth = levels.pixelHeight(level)
    else
        y = levels.pixelHeight(level) / 2
        bWidth = levels.TILE_SIZE * 2
    end

    local body = love.physics.newBody(level.physicsWorld,
      x, y, 0, 0)
    local shape = love.physics.newRectangleShape(
      body, 0, 0, bWidth, bHeight, 0 )
    shape:setRestitution(0)

    return {body, shape}
end

----------------------------------------------------------------------
-- 'map' is a global property used during level loading.  Global
-- because Tiled's Lua exporter is stupid.

local function loadLevelLayer(layer, width, index)
    for nx = 0, width - 1 do
        layer[nx] = {}
    end

    local maxTileID = 0

    local x
    local y = -1

    for i, tile in ipairs(map.layers[index].data) do
        if ((i - 1) % width) == 0 then
            x = 0
            y = y + 1
        end

        layer[x][y] = tile.gid

        x = x + 1
        
        if maxTileID < tile.gid then
            maxTileID = tile.gid
        end
    end

    return maxTileID
end

local function loadTerrain(level)
    level.groundLayer = {}

    local layerIndex = 1
    assert(map.layers[layerIndex].name == LAYER_GROUND)

    local maxTileID = loadLevelLayer(level.groundLayer, level.width,
      layerIndex)

    layerIndex = layerIndex + 1
    if map.layers[layerIndex].name == LAYER_OVER then

        level.overLayer = {}
        local newMaxTileID = loadLevelLayer(level.overLayer,
          level.width, layerIndex)

        if newMaxTileID > maxTileID then
            maxTileID = newMaxTileID
        end
    end

    return layerIndex, maxTileID
end

local function loadTileset(level, maxTileID)
    assert(#map.tilesets == 1)
    assert(#map.tilesets[1] > 1)

    local tilesetName = map.tilesets[1][1].source

    if not levels.TILESETS[tilesetName] then
        levels.TILESETS[tilesetName] = {
          image = graphics.loadImage(tilesetName, true),
          quads = {}
        }
    end

    local tilesetData = levels.TILESETS[tilesetName]

    if not tilesetData.walls then
        tilesetData.walls = {}

        for i = 2, #map.tilesets[1] do
            local tileType = map.tilesets[1][i]
            assert(tileType.label == 'tile')

            if tileType.properties.wall then
                tilesetData.walls[tileType.id + 1] = true
            end
        end
    end

    if #tilesetData.quads < maxTileID then
        local tilesetWidth = tilesetData.image:getWidth()
        local tilesetHeight = tilesetData.image:getHeight()

        local tilesetTileWidth = math.floor(
          tilesetWidth / levels.TILE_SIZE)

        for i = #tilesetData.quads + 1, maxTileID do
            local x, y = levels.indexToCoords(i, tilesetTileWidth)
            x = x * levels.TILE_SIZE
            y = y * levels.TILE_SIZE

            local quad = love.graphics.newQuad(x, y,
              levels.TILE_SIZE, levels.TILE_SIZE,
              tilesetWidth, tilesetHeight)

            table.insert(tilesetData.quads, quad)
        end
    end

    level.tileset = tilesetData
end

local function loadPhysics(level)
    local physicsWorld = love.physics.newWorld(
      levels.pixelWidth(level), levels.truePixelHeight(level))

    physicsWorld:setGravity(0, 0)
    physicsWorld:setMeter(levels.TILE_SIZE)

    level.physicsWorld = physicsWorld

    -- Block off level edges

    level.bounds = {
        makeBound(level, constants.DIR_CHANGES[constants.NORTH]),
        makeBound(level, constants.DIR_CHANGES[constants.EAST]),
        makeBound(level, constants.DIR_CHANGES[constants.SOUTH]),
        makeBound(level, constants.DIR_CHANGES[constants.WEST])
    }

    -- Wall bodies

    level.wallBodies = {}

    for x = 0, level.width - 1 do
        for y = 0, level.height - 1 do
            if levels.isWall(level, x, y) then
                local wallBody, wallShape = makeWall(level, x, y)

                -- Prevent walls from getting randomly deleted by
                -- garbage collector
                table.insert(level.wallBodies,
                  {body = wallBody, shape = wallShape} )
            end
        end
    end
end

local function loadActorData(level, layerIndex)
    if map.layers[layerIndex].name == LAYER_ACTORS then
        local x, y

        for _, data in ipairs(map.layers[layerIndex]) do
            x = data.x + levels.TILE_HALF_SIZE
            y = data.y + levels.TILE_HALF_SIZE

            if data.type == ACT_PLAYER then
                -- Place in center of tile
                level.playerSpawn = {x = x, y = y}
            else
                if not level.npcSpawn then
                    level.npcSpawn = {}
                end

                table.insert(level.npcSpawn,
                  {type = data.type, x = x, y = y} )
            end
        end

        layerIndex = layerIndex + 1
    end

    return layerIndex
end

local function loadPortalData(level, layerIndex)
    assert(map.layers[layerIndex].name == LAYER_PORTALS)

    level.inPortals = {}
    level.outPortals = {}

    local px, py
    local tx, ty
    local exitData

    for _, data in ipairs(map.layers[layerIndex]) do
        px = data.x + levels.TILE_HALF_SIZE
        py = data.y + levels.TILE_HALF_SIZE

        level.inPortals[data.type] = { x = px, y = py }

        exitData = { dest = data.properties[TRANS_TILE_LABEL] }
        if data.properties[TRANS_LEVEL_LABEL] then
            exitData.level = data.properties[TRANS_LEVEL_LABEL]
        end

        tx = data.x / levels.TILE_SIZE
        ty = data.y / levels.TILE_SIZE

        if not level.outPortals[tx] then
            level.outPortals[tx] = {}
        end

        level.outPortals[tx][ty] = exitData
        assert(level.outPortals[tx][ty])
    end
end

function levels.newLevel(file)
    local level = {}

    -- Create the global map object
    local levelChunk = love.filesystem.load(LEVEL_DIR .. file
      .. LEVEL_EXT)
    levelChunk()

    -- Load properties
    level.width = map.width
    level.height = map.height

    level.minY = 0
    if map.properties then
        level.minY = map.properties.minY
    end

    local layerIndex, maxTileID = loadTerrain(level)

    -- Load tileset after loading terrain, when maximum number of used
    -- tile types is known.
    loadTileset(level, maxTileID)

    loadPhysics(level)

    layerIndex = loadActorData(level, layerIndex + 1)
    loadPortalData(level, layerIndex)

--    level.physicsWorld:setCallbacks(addCollisionFunc, nil,
--      removeCollisionFunc)

    map = nil
    return level
end

----------------------------------------------------------------------

local function cameraRect(camera, level)
    local camTileX = levels.tile(camera.x)
    local camTileY = levels.tile(camera.y)

    local camTileWidth = math.ceil(camera.width / levels.TILE_SIZE)

    if ((camera.width % levels.TILE_SIZE) > 0)
      or (levels.offsetPixel(camera.x) > 0) then
        camTileWidth = camTileWidth + 1
    end

    local camTileHeight = math.ceil((camera.height + level.minY)
      / levels.TILE_SIZE)

    if ((camera.height % levels.TILE_SIZE) > 0)
      or (levels.offsetPixel(camera.y) > 0) then
        camTileHeight = camTileHeight + 1
    end

    return camTileX, camTileY, camTileWidth, camTileHeight
end

local function drawLayerRect(level, layer, minX, minY, maxX, maxY)
    for x = minX, maxX do
        for y = minY, maxY do
            local tileID = layer[x][y]
            if tileID > 0 then
                local quad = level.tileset.quads[tileID]
                love.graphics.drawq(level.tileset.image, quad,
                  x * levels.TILE_SIZE, y * levels.TILE_SIZE)
            end
        end
    end
end

local function drawLayer(level, layer, camera)
    local camTileX, camTileY, camTileWidth, camTileHeight
      = cameraRect(camera, level)

    -- Stay within map

    if camTileX < 0 then
        camTileWidth = camTileWidth + camTileX
        camTileX = 0
    end

    if camTileY < 0 then
        camTileHeight = camTileHeight + camTileY
        camTileY = 0
    end

    if (camTileX + camTileWidth) >= level.width then
        camTileWidth = level.width - camTileX
    end

    if (camTileY + camTileHeight) >= level.height then
        camTileHeight = level.height - camTileY
    end

    -- Draw the map
    drawLayerRect(level, layer, camTileX, camTileY,
      camTileX + camTileWidth - 1, camTileY + camTileHeight - 1)
end

function levels.drawGround(level, camera)
    drawLayer(level, level.groundLayer, camera)
end

function levels.drawOver(level, camera)
    if level.overLayer then
        drawLayer(level, level.overLayer, camera)
    end
end